# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import os
from pathlib import Path
from typing import Optional, Union

import pendulum as pend
from atomicfile import AtomicFile
from cached_property import cached_property

# pylint: disable=no-name-in-module
from fastnumbers import fast_int

from . import LOG, utils
from .clients import auto, base, local


class GotNoPostInfoError(Exception):
    pass


class Post:
    def __init__(self,
                 id_or_url: Union[None, int, str]   = None,
                 info:      Optional[base.InfoType] = None,
                 client:    Optional[base.Client]   = None
                ) -> None:
        assert id_or_url or info

        self.client = auto.get(client)

        if info:
            self.info = info

        elif isinstance(id_or_url, int):
            self.info = self.client.info_id(id_or_url)

        elif isinstance(id_or_url, (str, Path)):
            id_or_url   = id_or_url.strip()
            self.client = auto.get(id_or_url)
            self.info   = next(self.client.info_location(id_or_url))

        else:
            raise TypeError("id_or_url: must be int post ID or str URL.")

        if not self.info:
            raise GotNoPostInfoError(f"Got no info for post {id_or_url!r}.")

        if isinstance(info, local.IndexedInfo):
            return

        if "fetched_from" not in self.info:
            self.info["fetched_from"] = self.client.name

        if "fetched_at" not in self.info:
            self.info["fetched_at"] = pend.now()\
                                      .format(self.client.date_format)


    def __repr__(self) -> str:
        return f"{type(self).__name__}(id={self.id}, title='{self.title}')"


    def get_location(self, resource: str = "post", absolute: bool = False
                    ) -> str:
        return self.client.get_location(self.info, resource, absolute=absolute)


    @cached_property
    def title(self) -> str:
        kinds = {k: utils.join_comma_and(*self.info[f"tag_string_{k}"].split())
                 for k in ("character", "copyright", "artist")}

        return ("{character} ({copyright}) drawn by {artist}%".format(**kinds)
                .replace("() drawn by", "drawn by")
                .replace("drawn by %", "")
                .replace("%", "")
                .strip()
                or "untitled")

    @cached_property
    def id(self) -> int:
        return fast_int(self.info["id"])

    @cached_property
    def artcom(self) -> base.ArtcomType:
        return self.client.artcom(self.info)

    @cached_property
    def media(self) -> base.MediaType:
        return self.client.media(self.info)

    @cached_property
    def notes(self) -> base.NotesType:
        return self.client.notes(self.info)


    def download(self,
                 base_dir:  Union[str, Path] = Path("."),
                 overwrite: bool             = False,
                 warn:      bool             = True) -> None:

        if isinstance(self.client, local.Local):
            return

        post_dir = Path(base_dir) / "{fetched_from}-{id}".format(**self.info)
        post_dir.mkdir(parents=True, exist_ok=True)

        for res in ("info", "artcom", "notes", "media"):
            if res == "media" and "file_ext" not in self.info:
                LOG.warning("No decensor data found for post %d, "
                            "can't download media.", self.info["id"])
                continue

            ext = "json" if res != "media" else self.info["file_ext"]
            ext = "webm" if ext == "zip"   else ext

            path = post_dir / f"{res}.{ext}"

            if path.exists() and not overwrite:
                if warn:
                    LOG.warning("Not overwriting %r", str(path))
                continue

            content = getattr(self, res)

            if not content:
                continue

            if res != "media":
                with AtomicFile(path, "w") as out:
                    out.write("%s%s" %
                              (utils.jsonify(content).rstrip(), os.linesep))
                continue

            if self.info["file_ext"] != "zip":
                LOG.info("Downloading %s of %s for post %d",
                         self.info["file_ext"].upper(),
                         utils.bytes2human(self.info["file_size"]),
                         self.info["id"])
            else:
                LOG.info("Downloading WebM ugoira for post %d",
                         self.info["id"])

            with AtomicFile(path, "wb") as out:
                for chunk in content:  # pylint: disable=not-an-iterable
                    out.write(chunk)



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import collections
import time
from copy import copy
from pathlib import Path
from threading import Lock, Thread
from typing import List, Optional, Union

from dataclasses import dataclass, field

from . import LOG, config, order
from .clients import auto, base
from .filtering import filter_all
from .post import Post


@dataclass
class Stream(collections.Iterator):
    query:    Union[str, Path] = ""
    pages:    base.PageType    = 1
    limit:    Optional[int]    = None
    random:   bool             = False
    raw:      bool             = False
    client:   base.Client      = None
    location: bool             = False

    partial_tags:   bool = False
    filter_str:     str  = ""
    stop_if_filter: str  = ""

    unfinished:     List[Post] = field(init=False, default=None)
    posts_seen:     int        = field(init=False, default=0)
    downloaded:     int        = field(init=False, default=0)

    _info_gen: base.InfoGenType = \
        field(init=False, default=None, repr=False)

    _logged_iter_done: bool = field(init=False, default=False, repr=False)
    _applied_filters:  bool = field(init=False, default=False, repr=False)


    def __post_init__(self) -> None:
        self.client     = auto.get(self.client)
        self.unfinished = []

        if self.location or isinstance(self.query, Path):
            self._info_gen = self.client.info_location(self.query)
        else:
            self._info_gen = self.client.info_search(
                self.query, self.pages, self.limit, self.random, self.raw,
                partial_tags = True if self.partial_tags else False
            )

        auto_filter = config.CFG["GENERAL"]["auto_filter"]
        if not self.filter_str.startswith(auto_filter):
            self.filter_str = " ".join((self.filter_str, auto_filter)).strip()


    def _apply_filters(self) -> None:
        if self.filter_str:
            self._info_gen = filter_all(
                items        = self._info_gen,
                terms        = self.filter_str,
                raw          = self.raw,
                partial_tags = self.partial_tags
            )

        if self.stop_if_filter.strip():
            self._info_gen = filter_all(
                items         = self._info_gen,
                terms         = self.stop_if_filter,
                raw           = self.raw,
                stop_on_match = True,
                partial_tags  = self.partial_tags
            )


    def _on_iter_done(self, discarded: int) -> None:
        if self._logged_iter_done:
            return

        log = LOG.info if self.posts_seen else LOG.warning

        log("Found %d posts%s%s.",
            self.posts_seen,
            f", {discarded} filtered" if discarded  else "",
            f" for {self.query!r}"    if self.query else "")

        self._logged_iter_done = True


    def __next__(self) -> Post:
        # Don't do this at .__post_init__ to let .filter setup stuff before
        if not self._applied_filters:
            self._apply_filters()
            self._applied_filters = True

        while True:
            try:
                info = next(self._info_gen)
            except StopIteration as stop:
                self._on_iter_done(discarded=stop.value or 0)
                raise

            post = Post(info=info, client=self.client)

            self.posts_seen += 1
            return post


    def __iter__(self) -> "Stream":
        return self


    # pylint: disable=protected-access
    def filter(self, search: str, partial_tags: bool = False) -> "Stream":
        new              = copy(self)
        new.filter_str   = " ".join((search, new.filter_str)).strip()
        new.partial_tags = partial_tags
        return new

    def stop_if(self, search: str, partial_tags: bool = False) -> "Stream":
        new                = copy(self)
        new.stop_if_filter = " ".join((search, new.stop_if_filter)).strip()
        new.partial_tags   = partial_tags
        return new

    def order(self, by: str) -> "Album":
        from .album import Album  # ævoid circular dependency
        return Album(*order.sort(list(self), by))

    __truediv__  = lambda self, search: self.filter(search)        # /
    __floordiv__ = lambda self, search: self.filter(search, True)  # //
    __mod__      = lambda self, by:     self.order(by)             # %


    def download(self,
                 base_dir:  Union[str, Path] = Path("."),
                 overwrite: bool             = False,
                 warn:      bool             = True) -> "Stream":

        post        = None
        running     = {}
        thread_id   = 0
        max_running = int(config.CFG["GENERAL"]["parallel_requests"])
        lock        = Lock()

        def work(post: Post, thread_id: int) -> None:
            post.download(base_dir=base_dir, overwrite=overwrite, warn=warn)
            with lock:
                self.downloaded += 1
            del running[thread_id]

        try:
            while True:
                while len(running) >= max_running:
                    time.sleep(0.1)

                post = self.unfinished.pop(0) if self.unfinished else \
                       next(self)

                thread = Thread(target=work, args=(post, thread_id))
                running[thread_id] = thread
                thread.start()
                thread_id += 1

        except KeyboardInterrupt:
            LOG.warning("CTRL-C caught, finishing current tasks...")
            if post:
                self.unfinished.append(post)

        except StopIteration:
            pass

        return self



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import abc
import collections
import copy
import functools
from multiprocessing.pool import ThreadPool
from typing import Tuple

from . import LOG, config


class AttrIndexedDict(collections.UserDict, abc.ABC):
    "Dictionary where items are indexed by a specific attribute they possess."

    def __init_subclass__(cls,
                          attr:         str,
                          map_partials: Tuple[str] = (),
                          **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls.attr = attr

        for method in map_partials:
            setattr(cls, method, functools.partialmethod(cls.map, method))


    def __setitem__(self, key, value) -> None:
        raise RuntimeError("Use %s.put() to add items." % type(self).__name__)


    def __getattr__(self, name: str):
        "Allow accessing dict items with a dot like attributes."
        if name.startswith("_"):
            raise AttributeError(name)

        try:
            return self.data[name]
        except KeyError:
            raise AttributeError(f"No attribute or dict key named {name!r}.")


    def map(self, method: str, *args, _threaded: bool = True, **kwargs
           ) -> "AttrIndexedDict":
        "For all stored items, run a method they possess."

        work = lambda item: getattr(item, method)(*args, **kwargs)

        if _threaded:
            pool = ThreadPool(int(config.CFG["GENERAL"]["parallel_requests"]))

            try:
                pool.map(work, self.data.values())
            except KeyboardInterrupt:
                LOG.warning("CTRL-C caught, finishing current tasks...")
                pool.terminate()
            else:
                pool.close()

            pool.join()
            return self

        for item in self.data.values():
            work(item)
        return self


    def copy(self) -> "AttrIndexedDict":
        return copy.copy(self)


    def put(self, *items) -> "AttrIndexedDict":
        "Add items to the dict that will be indexed by self.attr."
        for item in items:
            self.data[getattr(item, self.attr)] = item
        return self

    # album << item, item >> album
    __lshift__ = __rrshift__ = lambda self, item: self.put(item)


    def merge(self, *others: "AttrIndexedDict") -> "AttrIndexedDict":
        for other in others:
            self.data.update(other)
        return self

    def __add__(self, other: "AttrIndexedDict") -> "AttrIndexedDict":
        return self.copy().merge(other)


    def difference(self, *others: "AttrIndexedDict") -> "AttrIndexedDict":
        new = self.copy()
        for other in others:
            for key in other:
                try:
                    del new[key]
                except KeyError:
                    pass
        return new

    def __sub__(self, other: "AttrIndexedDict") -> "AttrIndexedDict":
        return self.copy().difference(other)



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import math
import re
from typing import Any, Dict, Optional, Union
from urllib.parse import parse_qs, urlparse

import pendulum as pend
from dataclasses import dataclass, field

from pydecensooru import decensor, decensor_iter
# pylint: disable=no-name-in-module
from fastnumbers import fast_int

from . import base, net
from .. import LOG


@dataclass
class Danbooru(net.NetClient):
    name:      str = "danbooru"
    site_url:  str = "https://danbooru.donmai.us"
    username:  str = ""
    api_key:   str = ""

    default_limit: int = field(default=20,  repr=False)
    max_limit:     int = field(default=200, repr=False)

    url_templates: Dict[str, str] = field(default_factory=dict, repr=False)


    def __post_init__(self) -> None:
        super().__post_init__()

        self.url_templates = {
            "post":   "/posts/{id}",
            "artcom": "/artist_commentaries.json?search[post_id]={id}",
            "info":   "/posts/{id}.json",
            "notes":  "/notes.json?search[post_id]={id}",
        }

        net.ALIVE[self.name] = self


    def _api(self, endpoint_url: str, _catch_errs: bool = True, **params
            ) -> Union[None, Any]:
        response = self.http(
            "get",
            f"{self.site_url}/{endpoint_url}",
            params = params,
            auth   = ((self.username, self.api_key)
                      if self.username and self.api_key else None)
        )

        if not _catch_errs:
            return response.json()

        try:
            return response.json()
        except AttributeError:
            return []
        except ValueError as err:
            LOG.error(str(err))
            return []


    def info_id(self, post_id: int) -> Optional[base.InfoType]:
        info = self._api(f"posts/{post_id}.json")
        return decensor(info, self.site_url) if info else None


    def info_md5(self, md5: str) -> base.InfoGenType:
        yield from decensor_iter(self._api(f"posts.json", md5=md5),
                                 self.site_url)


    def info_search(self,
                    tags:   str           = "",
                    pages:  base.PageType = 1,
                    limit:  Optional[int] = None,
                    random: bool          = False,
                    raw:    bool          = False,
                    **kwargs) -> base.InfoGenType:

        # No need for other params if search is just an ID or MD5.
        if re.match(r"^id:\d+$", tags):
            post_id = fast_int(tags.split(":")[1], raise_on_invalid=True)
            LOG.info("Fetching post %d", post_id)
            info = self.info_id(post_id)
            if info:
                yield info
            return

        if re.match(r"^md5:[a-fA-F\d]+$", tags):
            LOG.info("Fetching post with MD5 %r", tags.split(":")[1])
            yield from self.info_md5(tags.split(":")[1])
            return

        params = {"tags": tags}

        if limit:
            params["limit"] = limit
            if limit > self.max_limit:
                LOG.warning("Max limit for %s is %d, got %d.",
                            self.name, self.max_limit, limit)

        # Do not pass "random=false", Danbooru sees it as true.
        if random is True:
            params["random"] = "true"

        if raw is True:
            params["raw"] = "true"

        total_posts = self.count_posts(params["tags"])
        last_page   = math.ceil(total_posts / (limit or self.default_limit))

        if total_posts == 0 or last_page == 0:
            LOG.warning("No posts for search %r.", tags)
            return

        fails = 0
        for page in self._parse_pages(pages, last_page):
            if page < 1:
                continue

            params["page"] = page

            LOG.info(
                "Fetching posts%s%s%s%s",
                " for %r"       % params["tags"] if params["tags"] else "",
                " on page %d%s" % (params["page"],
                                   f"/{last_page}" if last_page else ""),
                " [random]" if "random" in params else "",
                " [raw]"    if "raw"    in params else ""
            )

            try:
                yield from decensor_iter(
                    self._api("posts.json", **params, _catch_errs=False),
                    self.site_url
                )
            except AttributeError:
                fails += 1
            except ValueError as err:
                LOG.error(str(err))
                fails += 1
            else:
                fails = 0

            if fails >= 5:
                LOG.error("Giving up after 5 consecutive page fetch fails, "
                          "pagination limit probably reached.")
                return


    def info_location(self, location: str) -> base.InfoGenType:
        try:
            post_id = re.search(r"/posts/(\d+)\??.*$", location).group(1)
        except AttributeError:  # Not a direct post URL
            pass
        else:
            yield self.info_id(post_id)
            return

        parsed = parse_qs(urlparse(location).query)

        yield from self.info_search(
            tags   = parsed.get("tags",      [""]   )[-1],
            random = parsed.get("random",    [False])[-1],
            raw    = parsed.get("raw",       [False])[-1],
            pages  = fast_int(parsed.get("page",  [1]    )[-1]),
            limit  = fast_int(parsed.get("limit", [self.default_limit])[-1])
        )


    def artcom(self, info: base.InfoType) -> base.ArtcomType:
        if " commentary "         not in f" {info['tag_string_meta']} " and \
           " commentary_request " not in f" {info['tag_string_meta']} " and \
            pend.parse(info["created_at"]) > pend.yesterday():
            return []

        return self._api("artist_commentaries.json",
                         **{"search[post_id]": info["id"]})


    def media(self, info: base.InfoType) -> base.MediaType:
        if "file_ext" not in info:
            LOG.warning("No media available for post %d.", info["id"])
            return None

        url_key = "large_file_url" if info["file_ext"] == "zip" else "file_url"
        answer  = self.http("get", info[url_key], stream=True)

        try:
            return answer.iter_content(8 * 1024 ** 2)
        except AttributeError:
            return None


    def notes(self, info: base.InfoType) -> base.NotesType:
        if not bool(info["last_noted_at"]):
            return []

        return self._api("notes.json", **{"search[post_id]": info["id"]})


    def count_posts(self, tags: str = "") -> int:
        return self._api("counts/posts.json", tags=tags)["counts"]["posts"]


    def get_location(self, info: base.InfoType, resource: str = "post", **_
                    ) -> Optional[str]:
        assert resource in ("post", "artcom", "info", "media", "notes")

        if resource == "media" and "file_ext" not in info:
            LOG.warning("No decensor data found for post %d, "
                        "can't report media URL.", info["id"])
            return None

        if resource == "media" and info["file_ext"] == "zip":
            return info["large_file_url"]

        if resource == "media":
            return info["file_url"]

        return "%s%s" % (self.site_url,
                         self.url_templates[resource].format(**info))



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

"Misc useful functions."

import re
import sys
from typing import Union

import pendulum as pend
import simplejson
from colorama import Fore, Back, Style

# pylint: disable=no-name-in-module
from fastnumbers import fast_float, fast_int

SIZE_UNITS = "BKMGTPEZY"

def bytes2human(size: Union[int, float], prefix: str = "", suffix: str = ""
               ) -> str:
    size = fast_float(size)  # Prevent proxied size problems with round()
    unit = ""

    for unit in SIZE_UNITS:
        if size < 1024:
            break

        size /= 1024

    size = fast_int(size)    if unit in "BK" else \
           round(size, 1) if unit == "M" else \
           round(size, 2)

    return f"{size}{prefix}{unit}{suffix}"


def human2bytes(size: Union[int, float, str]) -> float:
    size   = str(size)
    result = fast_float(size.rstrip("%s%s" % (SIZE_UNITS, SIZE_UNITS.lower())))

    if size.lstrip(f"0123456789.").upper() in ("", "B", "O"):
        return result

    for unit in SIZE_UNITS.lstrip("B"):
        result *= 1024
        if size.rstrip("iIoObB")[-1].upper() == unit:
            break

    return result


def ratio2float(value: Union[int, float, str]) -> float:
    if isinstance(value, str) and ":" in value:
        w, h = value.split(":")
        return fast_int(w) / fast_int(h)

    return fast_float(value)


AGE_UNIT_TO_SUBTRACT_ARG = {
    ("Y", "y", "year"):                 "years",
    ("M", "mo", "month"):               "months",
    ("w", "week"):                      "weeks",
    ("d", "day"):                       "days",
    ("h", "hour"):                      "hours",
    ("m", "mi", "mn", "min", "minute"): "minutes",
    ("s", "sec", "second"):             "seconds",
    ("ms", "microsec", "microsecond"):  "microseconds"
}

def age2date(age: str) -> pend.DateTime:
    try:
        return pend.parse(age)  # If this is already a normal date
    except pend.parsing.exceptions.ParserError:
        pass

    user_unit  = age.lstrip("0123456789.")
    value      = abs(fast_float(age.replace(user_unit, "")))
    found_unit = None

    for units, shift_unit in AGE_UNIT_TO_SUBTRACT_ARG.items():
        for unit in units:
            if user_unit in (unit, f"{unit}s"):
                found_unit = shift_unit

    if not found_unit:
        raise ValueError(f"Invalid age unit: {user_unit!r}")

    return pend.now().subtract(**{found_unit: value})


JSONIFY_DEFAULT_PARAMS = {"sort_keys": True, "ensure_ascii": False}

def jsonify(dict_: dict, **dumps_kwargs) -> str:
    kwargs = {**JSONIFY_DEFAULT_PARAMS, **dumps_kwargs}
    return simplejson.dumps(dict_, **kwargs)


def join_comma_and(*strings: str) -> str:
    if len(strings) <= 1:
        return ", ".join(strings)

    return "%s and %s" % (", ".join(strings[:-1]), strings[-1])


def print_colored_help(doc: str, exit_code: int = 0) -> None:
    # ⋅ (U+22C5) is used as a docopt escape character
    doc = doc.replace("⋅", " ").splitlines()

    # Usage:
    doc[0] = re.sub(
        rf"(Usage: +)",
        rf"{Fore.MAGENTA}{Style.BRIGHT}\1{Style.RESET_ALL}{Fore.BLUE}",
        doc[0]
    )
    # [things]
    doc[0] = re.sub(rf"\[(\S+)\]",
                    rf"[{Style.BRIGHT}\1{Style.RESET_ALL}{Fore.BLUE}]",
                    doc[0])

    doc[0] = f"{doc[0]}{Fore.RESET}"
    doc    = "\n".join(doc)

    styles = {
        r"`(.+?)`":      ("GREEN",   "", ""),        # `things`
        r"^(\S.+:)$":    ("MAGENTA", "", "BRIGHT"),  #  Sections:
        r"^(  [A-Z]+)$": ("BLUE",    "", "BRIGHT"),  #  ARGUMENT
        r"^(  \S.+)$":   ("BLUE",    "", ""),        #  Two-space indented line
        r"^( {5,}-.+)$": ("BLUE",    "", ""),        #  Examples short-options
        r"^(\s*-)":      ("MAGENTA", "", ""),        #  - Dash lists
    }

    for regex, (fg, bg, style) in styles.items():
        fg    = getattr(Fore,  fg, "")
        bg    = getattr(Back,  bg, "")
        style = getattr(Style, style, "")
        doc   = re.sub(regex, rf"{fg}{bg}{style}\1{Style.RESET_ALL}", doc,
                       flags = re.MULTILINE)

    doc = re.sub(r"(-{1,2}[a-zA-Z\d]+ +)([A-Z]+)",
                 rf"\1{Fore.BLUE}{Style.BRIGHT}\2{Style.RESET_ALL}{Fore.BLUE}",
                 doc)

    print(doc)
    sys.exit(exit_code)



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import logzero
import colorama

colorama.init()

LOG = logzero.setup_logger(
    formatter = logzero.LogFormatter(
        datefmt = "%H:%M:%S",
        fmt     = ("%(color)s%(levelname)1.1s %(asctime)s | "
                   "%(message)s%(end_color)s"),
        colors  = {10: colorama.Fore.CYAN,   20: colorama.Fore.GREEN,
                   30: colorama.Fore.YELLOW, 40: colorama.Fore.RED,
                   50: colorama.Fore.MAGENTA}
    )
)

# pylint: disable=wrong-import-position

from .__about__ import __doc__
from . import config
config.reload()

from .post import Post
from .stream import Stream
from .album import Album



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import abc
import math
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union

import pendulum as pend
from dataclasses import dataclass, field

# pylint: disable=no-name-in-module
from fastnumbers import fast_int as fint

InfoType    = Union[Dict[str, Any], "IndexedInfo"]
InfoGenType = Generator[InfoType, None, None]

IE       = Union[int, type(Ellipsis)]
PageType = Union[IE, str, Tuple[IE, IE], Tuple[IE, IE, IE], Iterable[int]]

ArtcomType = NotesType = List[Dict[str, Any]]
MediaType  = Optional[Generator[bytes, None, None]]


@dataclass
class Client(abc.ABC):
    name: str = "client"

    date_format: str = \
        field(init=False, default="YYYY-MM-DDTHH:mm:ss.SSSZ", repr=False)


    @abc.abstractmethod
    def info_md5(self, md5: str) -> InfoGenType:
        yield {}


    @abc.abstractmethod
    def info_search(self,
                    tags:   str           = "",
                    pages:  PageType      = 1,
                    limit:  Optional[int] = None,
                    random: bool          = False,
                    raw:    bool          = False,
                    **kwargs) -> InfoGenType:
        yield {}


    @abc.abstractmethod
    def info_location(self, location: Union[str, Path]) -> InfoGenType:
        yield {}


    @abc.abstractmethod
    def artcom(self, info: InfoType) -> ArtcomType:
        return []


    @abc.abstractmethod
    def media(self, info: InfoType) -> MediaType:
        return None


    @abc.abstractmethod
    def notes(self, info: InfoType) -> NotesType:
        return []


    @abc.abstractmethod
    def count_posts(self, tags: str = "") -> int:
        return 0


    @abc.abstractmethod
    def get_location(self, info: InfoType, resource: str, **kwargs) -> str:
        return ""


    @staticmethod
    def _parse_pages(pages: PageType, last_page: int) -> Iterable[int]:
        is_str = isinstance(pages, str)

        if isinstance(pages, int) or (is_str and pages.isdigit()):
            return (fint(pages),)

        if pages in (..., "all"):
            return range(1, last_page + 1)

        if is_str and "-" in pages:
            pages = tuple(pages.split("-"))

        if isinstance(pages, tuple):
            begin = 1         if pages[0] in (..., "begin") else fint(pages[0])
            end   = last_page if pages[1] in (..., "end")   else fint(pages[1])
            step  = 1         if len(pages) < 3             else fint(pages[2])
            return range(begin, end + 1, step)

        if is_str and "," in pages:
            pages = [fint(p) for p in pages.split(",")]

        assert hasattr(pages, "__iter__"), "pages must be iterable"
        return pages


    @staticmethod
    def get_post_rank(post: "Post") -> int:
        score = post.info["score"]

        if score < 1:
            return score

        post_date = pend.parse(post.info["created_at"])

        if post_date > pend.now().subtract(days=2):
            return -1

        return math.log(score, 3) + post_date.timestamp() / 35_000



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import re
import shutil
from configparser import ConfigParser, ExtendedInterpolation
from pathlib import Path
from typing import Optional

from appdirs import user_config_dir
from pkg_resources import resource_filename

from . import __about__

DEFAULT_FILE = resource_filename(__about__.__name__, "data/default_config.ini")
FILE         = (Path(user_config_dir(__about__.__project_name__)) /
                f"{__about__.__pkg_name__}.ini")
CFG = ConfigParser(interpolation=ExtendedInterpolation())


def reload(path: Optional[str] = None) -> None:
    path = Path(path or FILE).expanduser()

    if not path.is_file():
        path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(DEFAULT_FILE, path)

    CFG.read_file(open(FILE, "r"))
    _reload_clients()


def _reload_clients() -> None:
    from . import clients  # Avoid circular import

    for name, cfg in CFG.items():
        if name in ("DEFAULT", "GENERAL"):
            continue

        assert re.match(r"[a-z0-9_-]+", name), \
               "Config section [{name}]: can only contain a-z 0-9 _ -"

        # Will be added to net.ALIVE (class __init__)
        client = clients.danbooru.Danbooru(
            site_url = cfg["site_url"], name    = name,
            username = cfg["username"], api_key = cfg["api_key"]
        )

        if name == CFG["GENERAL"]["default_booru"]:
            clients.net.DEFAULT = client



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import collections
import csv
import math
import multiprocessing as mp
import os
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Callable, Generator, List, Optional, Sequence, Union

import simplejson
from atomicfile import AtomicFile
from dataclasses import dataclass, field

# pylint: disable=no-name-in-module
from fastnumbers import fast_int

from . import base
from .. import LOG, order
from ..filtering import filter_all
from .base import InfoType


def str2bool(string: str) -> Union[bool, str, None]:
    return True  if string == "True"  else \
           False if string == "False" else \
           None  if not string        else string


def str2int(string: str) -> Optional[int]:
    return fast_int(string, None)


POST_FIELDS = collections.OrderedDict({
    # Starting keys to sort rows if needed
    "id":           str2int,
    "fetched_from": str,
    "fetched_at":   str,

    # Other keys useful for filtering
    "children_ids":           str,
    "created_at":             str,
    "fav_count":              str2int,
    "file_ext":               str,
    "file_size":              str2int,
    "image_height":           str2int,
    "image_width":            str2int,
    "is_deleted":             str2bool,
    "is_flagged":             str2bool,
    "is_note_locked":         str2bool,
    "is_pending":             str2bool,
    "is_rating_locked":       str2bool,
    "is_status_locked":       str2bool,
    "last_comment_bumped_at": str,
    "last_commented_at":      str,
    "last_noted_at":          str,
    "md5":                    str,
    "parent_id":              str2int,
    "rating":                 str,
    "score":                  str2int,
    "source":                 str,
    "tag_string":             str,
    "tag_string_artist":      str,
    "tag_string_character":   str,
    "tag_string_copyright":   str,
    "tag_string_general":     str,
    "tag_string_meta":        str,
    "updated_at":             str,
    "uploader_name":          str,

    # Keys not needed for filtering:
    "approver_id":          str2int,
    "bit_flags":            str2int,
    "down_score":           str2int,
    "file_url":             str,
    "has_active_children":  str2bool,
    "has_children":         str2bool,
    "has_large":            str2bool,
    "has_visible_children": str2bool,
    "is_banned":            str2bool,
    "is_favorited":         str2bool,
    "large_file_url":       str,
    "pixiv_id":             str2int,
    "pool_string":          str,
    "preview_file_url":     str,
    "tag_count":            str2int,
    "tag_count_artist":     str2int,
    "tag_count_character":  str2int,
    "tag_count_copyright":  str2int,
    "tag_count_general":    str2int,
    "tag_count_meta":       str2int,
    "up_score":             str2int,
    "uploader_id":          str2int,

    # Disabled for performance reasons
    # "keeper_data":             str2dict,
    # "pixiv_ugoira_frame_data": str2dict,
})


_INDEXED_INFO_NT = collections.namedtuple("IndexedInfo", POST_FIELDS.keys())

class IndexedInfo(_INDEXED_INFO_NT):
    _row_converts: List[Callable] = POST_FIELDS.values()


    @classmethod
    def from_csv(cls, row: Sequence[str]) -> "IndexedInfo":
        return cls(*(c(v) for c, v in zip(cls._row_converts, row)))


    def __getitem__(self, key):
        if isinstance(key, str):
            return getattr(self, key)
        return tuple(self)[key]


    def __contains__(self, key: str):
        return hasattr(self, key)


@dataclass
class Local(base.Client):
    name: str              = "local"
    path: Union[Path, str] = Path(".")

    index: Path = field(init=False, default=None, repr=False)


    def __post_init__(self) -> None:
        self.path  = Path(self.path or ".").expanduser()
        self.index = self.path / "index.tsv"


    def _get_info(self, post_dirname: str) ->  base.InfoType:
        path = self.path / post_dirname / "info.json"
        with open(path, "r") as file:
            return simplejson.loads(file.read())


    def _index_add(self, post_dirnames: List[str]) -> base.InfoGenType:
        LOG.info("Indexing %d posts...", len(post_dirnames))

        post_dirnames.sort(key     = lambda d: fast_int(d.split("-")[-1], -1),
                           reverse = True)

        if not self.index.exists():
            self.index.write_text("")

        with open      (self.index, "r", newline="") as in_file, \
             AtomicFile(self.index, "w")             as out_file:

            # Not just using csv.DictReader for performance reasons
            id_idx = list(POST_FIELDS.keys()).index("id")
            reader = csv.reader(in_file, delimiter="\t")

            src_row_writer = csv.writer(out_file, delimiter="\t")
            new_info_writer = csv.DictWriter(
                out_file,
                delimiter    = "\t",
                fieldnames   = POST_FIELDS.keys(),
                extrasaction = "ignore"
            )

            pool  = ThreadPool(mp.cpu_count() * 5)
            tasks = [pool.apply_async(self._get_info, (p,))
                     for p in post_dirnames]

            def info_gen():
                for task in tasks:
                    try:
                        yield task.get()
                    except (FileNotFoundError, NotADirectoryError) as err:
                        if str(err.filename) != self.index.name:
                            LOG.error(str(err))

            info_gen = info_gen()
            try:
                new_info = next(info_gen)
            except StopIteration:
                return

            no_more_to_add = False

            for source_row in reader:
                try:
                    src_id = fast_int(source_row[id_idx],raise_on_invalid=True)
                except ValueError:
                    LOG.error("Removing invalid row in index: %r", source_row)
                    continue

                while not no_more_to_add and new_info["id"] > src_id:
                    new_info_writer.writerow(new_info)
                    yield new_info

                    try:
                        new_info = next(info_gen)
                    except StopIteration:
                        no_more_to_add = True

                src_row_writer.writerow(source_row)

            if not no_more_to_add:
                new_info_writer.writerow(new_info)
                yield new_info

            for remaining_info in info_gen:
                new_info_writer.writerow(remaining_info)
                yield remaining_info


    def _index_del(self, *line_nums: int) -> None:
        LOG.info("Deleting from index %d post dirs...", len(line_nums))

        with open      (self.index, "r", newline="") as in_file, \
             AtomicFile(self.index, "w")             as out_file:

            for i, line in enumerate(in_file, 1):
                if i not in line_nums:
                    out_file.write(line)


    def _index_iter(self, post_dirnames: List[str]
                   ) -> Generator[IndexedInfo, None, None]:

        if not self.index.exists():
            yield from self._index_add(post_dirnames)
            return

        lines_to_del     = []
        unfound_dirnames = set(post_dirnames)
        unfound_dirnames.discard(self.index.name)
        del post_dirnames

        with open(self.index, "r", newline="") as file:
            reader = csv.reader(file, delimiter="\t")

            for i, row in enumerate(reader, 1):
                try:
                    info = IndexedInfo.from_csv(row)
                except TypeError:
                    LOG.error("Corrupted post in index on line %d, "
                              "will try to repair: %r", i, row)
                    lines_to_del.append(i)
                    continue

                key  = f"{info.fetched_from}-{info.id}"

                try:
                    unfound_dirnames.remove(key)
                except KeyError:
                    lines_to_del.append(i)
                else:
                    yield info

        if lines_to_del:
            self._index_del(*lines_to_del)

        if unfound_dirnames:
            yield from self._index_add(list(unfound_dirnames))


    def _get_post_path(self, info: InfoType) -> Path:
        return self.path / f"{info['fetched_from']}-{info['id']}"


    def _read_res(self, info: InfoType, file: str, binary: bool = False
                 ) -> Union[str, bytes, None]:

        path = self._get_post_path(info)

        try:
            path /= file
        except FileNotFoundError:
            return None

        return path.read_text() if not binary else path.read_bytes()


    def _read_json(self, info: InfoType, file_noext: str) -> Optional[str]:
        res = self._read_res(info, f"{file_noext}.json")
        if res is None:
            return None
        return simplejson.loads(res)


    def info_id(self, post_id: str) -> Optional[base.InfoType]:
        try:
            _, _ = post_id.split("-")
            fast_int(post_id, raise_on_invalid=True)
        except (AttributeError, ValueError):
            raise ValueError("post_id for local clients must be in the form "
                             "'<booru>-<id>', e.g. 'danbooru-1'.")


        path = self.path / post_id / "info.json"
        try:
            return simplejson.loads(path.read_text())
        except FileNotFoundError:
            return None


    def info_md5(self, md5: str) -> base.InfoGenType:
        yield from self.info_search(f"md5:{md5}")


    # pylint: disable=arguments-differ
    def info_search(self,
                    tags:         str           = "",
                    pages:        base.PageType = 1,
                    limit:        Optional[int] = None,
                    random:       bool          = False,
                    raw:          bool          = False,
                    partial_tags: bool          = False) -> base.InfoGenType:

        posts = os.listdir(self.path)

        ok_i = max_i = None

        if limit and limit not in (-1, math.inf):
            last  = math.ceil(len(posts) / limit)
            ok_i  = {i
                     for p in self._parse_pages(pages, last)
                     for i in range((p-1) * limit, (p-1) * limit + limit)}
            max_i = sorted(ok_i)[-1]

        posts = filter_all(self._index_iter(posts),
                           terms=tags, raw=raw, partial_tags=partial_tags)

        if random:
            posts = iter(order.sort(list(posts), by="random"))

        for i, post in enumerate(posts):
            if max_i and i > max_i:
                break

            if ok_i and i not in ok_i:
                continue

            yield post


    def info_location(self, location: Union[str, Path]) -> base.InfoGenType:
        path = Path(location).expanduser()
        read = lambda info_path: simplejson.loads(info_path.read_text())

        if path.name == "info.json":
            yield read(path)
            return

        if path.name in ("artcom.json", "notes.json") or path.stem == "media":
            yield read(path.parent / "info.json")
            return

        try:
            yield read(path / "info.json")
            return
        except FileNotFoundError:
            pass

        if path.resolve() == self.path.resolve():
            yield from self.info_search()
            return

        raise ValueError(
            f"'{path!s}': location (path) must be a post dir, "
            f"file inside a post dir or the dir containing all the posts."
        )


    def artcom(self, info: InfoType) -> base.ArtcomType:
        return self._read_json(info, "artcom") or []


    def media(self, info: InfoType) -> base.MediaType:
        try:
            ext = "webm" if info["file_ext"] == "zip" else info["file_ext"]
        except KeyError:
            return None
        return self._read_res(info, f"media.{ext}", binary=True)


    def notes(self, info: InfoType) -> base.NotesType:
        return self._read_json(info, "notes") or []


    def count_posts(self, tags: str = "") -> int:
        return len(list(self.info_search(tags)))


    def get_location(self,
                     info:     base.InfoType,
                     resource: str  = "post",
                     absolute: bool = False) -> Optional[str]:

        def verify(path: Path) -> Optional[str]:
            if path.exists():
                return str(path)
            return None

        assert resource in ("post", "artcom", "info", "media", "notes")
        path = self._get_post_path(info)

        if absolute:
            path = path.resolve()

        if resource == "post":
            return verify(path)

        if resource == "media":
            try:
                ext = "webm" if info["file_ext"] == "zip" else info["file_ext"]
            except KeyError:
                return None

            return verify(path / f"media.{ext}")

        return verify(path / f"{resource}.json")



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import random
from typing import List

import pendulum as pend

# pylint: disable=no-name-in-module
from fastnumbers import fast_int

from .post import Post

ORDER_NUM = {
    "id":       ("asc",  "id"),
    "score":    ("desc", "score"),
    "favcount": ("desc", "fav_count"),
    "filesize": ("desc", "file_size"),
    "tagcount": ("desc", "tag_count"),  # TODO: document on dan wiki
    "gentags":  ("desc", "tag_count_general"),
    "arttags":  ("desc", "tag_count_artist"),
    "chartags": ("desc", "tag_count_character"),
    "copytags": ("desc", "tag_count_copyright"),
    "metatags": ("desc", "tag_count_meta"),
    "mpixels":  ("desc",
                 lambda i: (fast_int(i["image_width"],  0) *
                            fast_int(i["image_height"], 0)) / 1_000_000),
    # Non-standard:
    "width":  ("desc", "image_width"),
    "height": ("desc", "image_height"),
}

ORDER_DATE = {
    "change":         ("desc", "updated_at"),
    "comment":        ("desc", "last_commented_at"),
    "comm":           ("desc", "last_commented_at"),
    "comment_bumped": ("desc", "last_commented_bumped_at"),
    "note":           ("desc", "last_noted_at"),
    # Non-standard:
    "created": ("desc", "created_at"),
    "fetched": ("desc", "fetched_at"),
}

ORDER_FUNCS = {
    # pylint: disable=unnecessary-lambda
    "rank":      lambda p: p.info.client.get_post_rank(p),
    "random":    lambda _: random.random(),
    "landscape": lambda p: int(p.info["image_width"] > p.info["image_height"]),
    "portrait":  lambda p: int(p.info["image_height"] > p.info["image_width"]),
}


def sort(posts: List[Post], by: str) -> List[Post]:
    by_val  = by.replace("asc_", "").replace("desc_", "")

    in_dict = (ORDER_NUM   if by_val in ORDER_NUM   else
               ORDER_DATE  if by_val in ORDER_DATE  else
               ORDER_FUNCS if by_val in ORDER_FUNCS else None)

    if not in_dict:
        raise ValueError(
            f"Got {by_val!r} as ordering method, must be one of: %s" %
            ", ".join(set(ORDER_NUM) | set(ORDER_DATE) | set(ORDER_FUNCS))
        )

    if in_dict == ORDER_FUNCS:
        posts.sort(key=ORDER_FUNCS[by], reverse=(by != "random"))
        return posts

    by_full = by if by.startswith("asc_") or by.startswith("desc_") else \
              f"%s_{by}" % in_dict[by][0]

    def sort_key(post: Post) -> int:
        key = in_dict[by_val][1]
        key = post.info[key] if not callable(key) else key(post.info)
        return pend.parse(key) if in_dict == ORDER_DATE else key

    posts.sort(key=sort_key, reverse=by_full.startswith("desc_"))
    return posts



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

"Search, filter, inspect, download posts from Danbooru and local folders."

__project_name__ = "lunakit"
__pkg_name__     = "lunafind"
__version__      = "0.5.4-2"
__status__       = "Development"
# __status__     = "Production"

__author__ = "miruka"
__email__  = "miruka@disroot.org"

__license__ = "LGPLv3"



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import traceback
from typing import Generator, List

from . import LOG, order
from .attridict import AttrIndexedDict
from .filtering import filter_all
from .post import Post
from .stream import Stream


class Album(AttrIndexedDict, attr="id", map_partials=("download",)):
    def __init__(self, *stream_args_posts_streams, **stream_kwargs) -> None:
        super().__init__()
        self._added: int = 0
        self.put(*stream_args_posts_streams, **stream_kwargs)


    def __repr__(self) -> str:
        return "%s({\n%s\n})" % (
            type(self).__name__,
            "\n".join((f"  {i!r}: '{p.title:.64}'," for i, p in self.items()))
        )


    @property
    def list(self) -> List[Post]:
        return list(self.values())


    def _put_post(self, post: Post) -> None:
        try:
            super().put(post)
            self._added += 1
        except StopIteration:
            raise
        except Exception:
            traceback.print_exc()
            LOG.error("Unexpected error while handling post %d, "
                      "trying to recover...", post.id)

    def _put_stream(self, stream: Stream) -> None:
        for post in stream:
            self._put_post(post)

    # pylint: disable=arguments-differ
    def put(self, *stream_args_posts_streams, **stream_kwargs) -> "Album":
        stream_args = []
        self._added = 0

        try:
            for arg in stream_args_posts_streams:
                if isinstance(arg, Post):
                    self._put_post(arg)

                elif isinstance(arg, Stream):
                    self._put_stream(arg)

                else:
                    stream_args.append(arg)

            if stream_args or stream_kwargs:
                self._put_stream(Stream(*stream_args, **stream_kwargs))

        except KeyboardInterrupt:
            LOG.warning("Caught CTRL-C, added %d posts.", self._added)

        return self


    def filter(self, search: str, partial_tags: bool = False) -> "Album":
        return Album(*self.filter_lazy(search, partial_tags))

    def filter_lazy(self, search: str, partial_tags: bool = False
                   ) -> Generator[Post, None, None]:
        yield from filter_all(self.list, search, partial_tags=partial_tags)

    def order(self, by: str) -> "Album":
        return Album(*order.sort(self.list, by))

    __truediv__  = lambda self, search: self.filter(search)        # /
    __floordiv__ = lambda self, search: self.filter(search, True)  # //
    __mod__      = lambda self, by:     self.order(by)             # %



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

"lunafind setuptools file"

from setuptools import setup, find_packages

from lunafind import __about__


def get_readme():
    with open("README.md", "r") as readme:
        return readme.read()


setup(
    name        = __about__.__pkg_name__,
    version     = __about__.__version__,

    author       = __about__.__author__,
    author_email = __about__.__email__,
    license      = __about__.__license__,

    description                   = __about__.__doc__,
    long_description              = get_readme(),
    long_description_content_type = "text/markdown",

    python_requires  = ">=3.6, <4",
    install_requires = [
        "appdirs",
        "atomicfile",
        "cached_property",
        "colorama",
        "dataclasses;python_version<'3.7'",
        "docopt",
        "fastnumbers",
        "logzero",
        "pendulum",
        "pydecensooru>=0.1.0",
        "requests",
        "setuptools",
        # Needed because simplejson converts namedtuples to dicts unlike json
        "simplejson",
        "urllib3",
        "whratio>=3.1.1",
    ],

    include_package_data = True,
    package_data         = {__about__.__pkg_name__: ["data/*"]},
    packages             = find_packages(),
    entry_points    = {
        "console_scripts": [
            f"{__about__.__pkg_name__}={__about__.__pkg_name__}.cli:main"
        ]
    },

    keywords = "lunakit lunasync booru danbooru api client image ugoira " \
               "anime cli terminal scrap tag search filter order download",

    url = "https://github.com/mirukan/lunafind",

    classifiers=[
        "Development Status :: 3 - Alpha",
        # "Development Status :: 4 - Beta",
        # "Development Status :: 5 - Production/Stable",

        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",

        "Environment :: Console",

        "Topic :: Utilities",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Software Development :: Libraries :: Python Modules",

        ("License :: OSI Approved :: "
         "GNU Lesser General Public License v3 or later (LGPLv3+)"),

        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",

        "Natural Language :: English",

        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
    ]
)



import re
from pathlib import Path
from typing import Union

from . import base
from .. import LOG


class NoClientFoundError(Exception):
    def __init__(self, premsg: str) -> None:
        super().__init__(
            f"{premsg} - Initialize a NetClient object, or add an appropriate "
            f"section in the lunafind config file."
        )


def get(value: Union[str, base.Client, None]) -> base.Client:
    from . import net

    if value is None:
        return net.DEFAULT

    if isinstance(value, base.Client):
        return value

    if isinstance(value, str) and value in net.ALIVE:
        return net.ALIVE[value]

    if isinstance(value, str):
        value = value.strip()


    if isinstance(value, str) and re.match(r"^\s*https?://.+", value):
        for name, client in net.ALIVE.items():
            site = client.site_url

            if value.startswith("http://") and site.startswith("https://"):
                value = value.replace("http://", "https://")

            if value.startswith(site):
                LOG.info("Auto-detect: using net client for %r.", name)
                return client

        raise NoClientFoundError(f"No appropriate client found for "
                                 f"URL {value!r}.")

    # In last resort, assume value is a path
    path = Path(value).expanduser()

    if not path.exists():
        raise FileNotFoundError("Path %r doesn't exist." % str(path))

    if path.is_file():
        path = path.parent

    if (path / "info.json").exists():
        path = path.parent

    LOG.info("Auto-detect: using local client for %s.",
             f"directory '{path!s}'" if path != Path(".") else "current dir")

    from . import local
    return local.Local(path)



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

import re
import shlex
from typing import Generator, Iterable, Optional, Set, Union

import pendulum as pend

import whratio
# pylint: disable=no-name-in-module
from fastnumbers import fast_float, fast_int

from . import LOG, utils
from .clients.base import InfoType
from .post import Post


def parse_date(value) -> pend.DateTime:
    # tz="local" only applies if there's no tz in the value
    return pend.parse(value, tz="local")


def safe_int(value) -> int:
    return fast_int(value, 0)


META_NUM_TAGS = {
    "width":    ["image_width",         safe_int],
    "height":   ["image_height",        safe_int],
    "score":    ["score",               safe_int],
    "favcount": ["fav_count",           safe_int],
    "id":       ["id",                  safe_int],
    "pixiv":    ["pixiv_id",            safe_int],
    "tagcount": ["tag_count",           safe_int],
    "gentags":  ["tag_count_general",   safe_int],
    "arttags":  ["tag_count_artist",    safe_int],
    "chartags": ["tag_count_character", safe_int],
    "copytags": ["tag_count_copyright", safe_int],
    "metatags": ["tag_count_meta",      safe_int],
    "date":     ["created_at",          parse_date],

    "mpixels": [lambda i: (safe_int(i["image_width"]) *
                           safe_int(i["image_height"])) / 1_000_000,
                fast_float],

    # Non-standard addition: support both float and int:int user ratios.
    "ratio":    [lambda i: whratio.as_float(fast_int(i["image_width"],  1),
                                            fast_int(i["image_height"], 1)),
                 utils.ratio2float],

    # Non-standard addition: supports microseconds and more units aliases.
    "age": ["created_at",  utils.age2date, "reverse_cmp"],
    # none, any, or the post number the post should be a child of.
    "parent": ["parent_id", fast_int],
    # none, any, or (non-standard) a number of possessed children.
    "child": [lambda i: len((i["children_ids"] or "").split()), safe_int],
    # Non-standard addition: supports units other than b/KB/MB
    "filesize": ["file_size", utils.human2bytes, "eq_fuzzy_20"],

    "fetch":    ["fetched_at", parse_date],
    "fetchage": ["fetched_at", utils.age2date, "reverse_cmp"],
}


def _source_match(info: InfoType, value: str, key: str = "source") -> bool:
    if value == "none":
        return not info[key]

    if value.startswith("pixiv/"):
        # If no / at the end, will match artists *starting with* value.
        return re.search(rf"pixiv(\.net/img.*(/img)?)?/{value[6:]}",
                         info[key], re.IGNORECASE)

    # Other "source:..." on Danbooru: "match anything that starts with ...".
    # * in value = .* regex (other regexes get escaped).
    return re.match(r"%s.*" % re.escape(value).replace(r"\*", r".*"),
                    info[key], re.IGNORECASE)


META_STR_TAGS_FUNCS = {
    "md5":      lambda info, v: info["md5"]          == v,
    "filetype": lambda info, v: info.get("file_ext") == v,
    "rating":   lambda info, v: info["rating"][0]    == v[0],
    "locked":   lambda info, v: info[f"is_{v}_locked"],
    "status":   lambda info, v: v in ("any", "all") or info[f"is_{v}"],
    "source":   _source_match,

    # Non-standard tags:
    "from": lambda info, v: info["fetched_from"] == v,
}


def _tag_present(info: InfoType, tag: str) -> bool:
    # Avoid regex unless no choice, they have bad performances.
    # Wrap strings in spaces to match tags even if they're at start/end.
    if "*" not in tag:
        return f" {tag} " in f" {info['tag_string']} "

    if tag == "*":
        return True

    if "*" in tag and "*" not in tag[1:-1]:
        tag = f"{tag[1:]}"  if tag[0]  == "*" else f" {tag}"
        tag = f"{tag[:-1]}" if tag[-1] == "*" else f"{tag} "
        return tag in f" {info['tag_string']} "

    # Non-standard: support wildcards in "-tag" or "~tag".
    # "*" in tag → ".*?" regex, escape other regex/special chars
    return re.search(r" %s " % re.escape(tag).replace(r"\*", r".*?"),
                     r" %s " % info["tag_string"],
                     re.IGNORECASE)


def _meta_num_match(info: InfoType, tag: str, value: str) -> bool:
    convert = META_NUM_TAGS[tag][1]

    info_v = META_NUM_TAGS[tag][0]
    info_v = convert(
        info_v(info) if callable(META_NUM_TAGS[tag][0]) else info[info_v]
    )

    eq_fuzzy_20 = "eq_fuzzy_20" in META_NUM_TAGS[tag]
    reverse_cmp = "reverse_cmp" in META_NUM_TAGS[tag]

    def compare(convert, info_v, value, eq_fuzzy_20, reverse_cmp) -> bool:
        # Non-standard: none/any supported for all meta numeric tags.
        if value == "none":
            return bool(not info_v)

        if value == "any":
            return bool(info_v)

        if value.startswith(">="):
            return info_v >= convert(value[2:])
        if value.endswith(".."):
            return info_v >= convert(value[:-2])

        if value.startswith("<=") or value.startswith(".."):
            return info_v <= convert(value[2:])

        if value.startswith(">"):
            return info_v > convert(value[1:])

        if value.startswith("<"):
            return info_v < convert(value[1:])

        if ".." in value:
            begin, end = value.split("..", maxsplit=1)

            if reverse_cmp:
                begin, end = end, begin
                return not convert(begin) <= info_v <= convert(end)

            return convert(begin) <= info_v <= convert(end)

        if "," in value:
            if not eq_fuzzy_20:
                return str(info_v) in value.split(",")

            return bool([v for v in value.split(",") if
                         v - v / 20 <= info_v <= v + v / 20])


        try:
            value = convert(value)

            if eq_fuzzy_20:
                # For filesize, it's what Danbooru does apparently
                # (not sure if this is the exact formula but close enough?).
                return value - value / 20 <= info_v <= value + value / 20

            return info_v == value
        except Exception:
            pass

        raise ValueError(f"Invalid search term value: '{tag}:{value}'.")

    result = compare(convert, info_v, value, eq_fuzzy_20, reverse_cmp)
    return result if not reverse_cmp else not result


def _filter_info(info:        InfoType,
                 simple_tags: Set[str],
                 meta_num:    Set[str],
                 meta_str:    Set[str]) -> bool:

    no_prefix = lambda tag: tag[1:] if tag[0] in ("-", "~") else tag
    presences    = {}

    for tag in simple_tags:
        presences[tag] = _tag_present(info, no_prefix(tag))

    for tag_val in meta_num:
        tag, value         = tag_val.split(":", maxsplit=1)
        presences[tag_val] = _meta_num_match(info, no_prefix(tag), value)

    for tag_val in meta_str:
        tag, value         = tag_val.split(":", maxsplit=1)
        presences[tag_val] = META_STR_TAGS_FUNCS[no_prefix(tag)](info, value)

    tilde_tag_in_search   = False
    one_tilde_tag_present = False

    for term, present in presences.items():
        if term[0] == "-" and present:
            return False

        if term[0] not in ("-", "~") and not present:
            return False

        if term[0] == "~":
            tilde_tag_in_search = True

            if present:
                one_tilde_tag_present = True

    if tilde_tag_in_search and not one_tilde_tag_present:
        return False

    return True


def filter_all(items:         Iterable[Union[InfoType, Post]],
               terms:         str,
               raw:           bool = False,
               stop_on_match: bool = False,
               partial_tags:  bool = False,
              ) -> Generator[Union[InfoType, Post], None, int]:

    def raw_tag(term: str) -> Optional[str]:
        if not ":" in term:
            return None
        tag = term.split(":")[0]
        return tag[1:] if tag[0] in ("-", "~") else tag

    terms     = set(shlex.split(terms.lower())) if not raw else {terms.lower()}
    meta_num  = set(t for t in terms if raw_tag(t) in META_NUM_TAGS)
    meta_str  = set(t for t in terms if raw_tag(t) in META_STR_TAGS_FUNCS)
    tags      = terms - set(meta_num) - set(meta_str)

    if partial_tags:
        tags = {re.sub(r"^(-|~)?(.+)", r"\1*\2*", t) for t in tags}

    term_args = (tags, meta_num, meta_str)

    discarded = 0

    for item in items:
        info = item.info if isinstance(item, Post) else item

        try:
            if stop_on_match:
                if not _filter_info(info, *term_args):
                    yield item
                else:
                    return 0

            else:
                if _filter_info(info, *term_args):
                    yield item
                else:
                    discarded += 1

        except KeyError as err:
            LOG.warning("No %r key for post %d.", err.args[0], info["id"])

    return discarded



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

"Base network client abstract class and other net-related variables."

import abc
import threading
from typing import Dict, Optional

import urllib3
from dataclasses import dataclass, field

import requests
from requests.adapters import HTTPAdapter

from . import base
from .. import LOG, config

# Set the maximum number of total requests/threads that can be running at once.
# When doing Album.write(), a .write() thread will launch for every Post,
# and every Post will launch a .write() thread for every resource, etc.
MAX_PARALLEL_REQUESTS_SEMAPHORE = threading.BoundedSemaphore(
    int(config.CFG["GENERAL"]["parallel_requests"])
)

RETRY = urllib3.util.Retry(
    total                      = 4,
    redirect                   = 8,
    status_forcelist           = [420, 421, 429, 500, 502, 503],
    backoff_factor             = 1.5,
    raise_on_redirect          = False,
    raise_on_status            = False,
    respect_retry_after_header = True
)


ALIVE:   Dict[str, "NetClient"] = {}
DEFAULT: Optional["NetClient"]  = None


# pylint: disable=abstract-method
@dataclass
class NetClient(base.Client, abc.ABC):
    name:     str = "netclient"
    site_url: str = ""

    _session: requests.Session = field(init=False, default=None, repr=False)


    def __post_init__(self) -> None:
        self._session = requests.Session()

        for scheme in ("http://", "https://"):
            self._session.mount(scheme, HTTPAdapter(max_retries=RETRY))


    def http(self, http_method: str, url: str, **request_kwargs
            ) -> Optional[requests.models.Response]:
        try:
            with MAX_PARALLEL_REQUESTS_SEMAPHORE:
                response = self._session.request(
                    http_method, url, timeout=6.5, **request_kwargs
                )
                response.raise_for_status()

        except requests.exceptions.RequestException as err:
            LOG.error(str(err))
            return None

        return response


    @abc.abstractmethod
    def info_id(self, post_id: int) -> Optional[base.InfoType]:
        return None



# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

r"""Usage: lunafind [QUERY]... [options]

Search, filter and download posts from Danbooru-based sites.
Local directories containing downloaded posts can also be indexed and searched.

Part of the lunakit tools.
See also `lunasync` to easily download and keep in sync particular tags,
similar to Danbooru subscriptions: `https://github.com/mirukan/lunasync`

Arguments:
  QUERY
    Tags to search for. If unspecified, latest posts will be returned.

    Use the `-q`/`--query-location` option to take a search results URL,
    post URL, post resource path, or post directory path instead.

    As multiple queries can be used in one command,
    remember to quote searches with spaces (e.g. multiple tags).

    If a query starts with a `-`, prefix it with a `%` or `\` to prevent it
    from being seen as an option, e.g. `lunafind %-rating:e`.
    For `\`, quoting will always be required to prevent the shell from
    interpreting it.

Options:
  -q, --query-location
    Take booru search results URL, post URL, local post directory or
    resource path as `QUERY` instead of search tags.

  -p PAGES, --pages PAGES
    Pages to fetch for tag searches, can be:
    - All pages from start to end: `all`
    - A direct number for a single page, e.g. `1` (the default)
    - A range with an optional step: `3-6`, `10-end`, `1-20-2` (skip 1/2 pages)
    - A comma-separated list of pages: `1,3,5,12,99`

  -l NUM, --limit NUM
    Number of posts per page.

    For Danbooru, default is `20`, max is `200`, and the max page with the
    default limit for non-premium users is `1000`.
    Using `--limit 200` allows more total posts to be fetched,
    20K vs 80K with some big tags. Timeouts start to appear near pages 400-500.
    After 5 fatal timeouts (which are each retried a few times), the search
    will give up and return what was found until then.

    The default for local directories is infinite (`-1`), so the `-p`/`--pages`
    option will have no special effect unless a limit is set.

  -r, --random
    Get results in a randomized order.
    For local posts, all results have to be loaded in RAM in order to be
    effectively randomized, this could take some time.

  -w, --raw
    Do not parse the query for aliased tags, metatags or multiple tags,
    send it as a single literal tag.

  -s SOURCE, --source SOURCE
    Where to search posts.
    Can be the name of a booru defined in your configuration file,
    or the path to a folder containing downloaded posts.

    If a relative folder with the same name as a booru is used as `SOURCE`,
    e.g. `danbooru`, use `./danbooru` instead to indicate that it's a path.

    If not specified, the default booru from your config file is used.
    This option is ignored for URL/path queries with `q`/`--query-location`.


  -f TAGS, --filter TAGS
    Filter posts returned by searches,
    can be used to work around the two tags limit on Danbooru.
    Same syntax as tag searches, most metatags are supported with some
    additions and restriction lifts (to document).

    For a faster filtering, use tags that have less posts for the booru
    search and others for the filter, for example:
      `lunafind "snow wallpaper" -f "touhou 1girl" -D`
    Instead of:
      `lunafind "touhou 1girl" -f "wallpaper snow" -D`

  -m, --partial-match
    For `--filter` tags or `--source local` search queries,
    make every tag act like they are surrounded by wildcards, e.g.:
      `lunafind red -m -s local`
    would return results for tags like `red`, `red_eyes`, `bored`, etc, instead
    of just `red`.

    This allows, for example, to type only a character's first name and still
    get expected results.

  -o BY, --order BY
    Order posts returned by searches.
    Has to force all results to be fetched at the start and loaded in RAM.
    See `--help-order-values` to get a list of the possible `BY` values.

    Also remember this works with actually fetched results.
    The equivalent to searching "tag1 tag2 order:score" on Danbooru would be:
    `lunafind "tag1 tag2" --pages all --order score` (notice `--pages all`).


  -R RESOURCE, --resource RESOURCE
    Posts resource to print on stdout, default is `info`.
    Can be `info`, `media`, `artcom` or `notes`.

  -S RESOURCE, --show-location RESOURCE
    If `RESOURCE` is `post`: print the URL or directory path of posts.
    If `RESOURCE` is `info`, `media`, `artcom` or `notes`:
    print the URL or file path of that resource for posts.

  -A, --absolute-path
    Print absolute paths for `-S`/`--show-location`, resolving any symlinks.
    This can be a lot slower than printing normal relative paths.

  -D DIR, --download DIR
    Save posts and their resources (media, info, artcom, notes) to a folder.
    With `-D .` for example, info for post 1 on Danbooru would be saved
    at `./danbooru-1/info.json`.
    Any missing directory is created.
    Has no effect for posts from local directories.

    Do not rename the downloaded post directories (`<booru>-<id>`),
    or they will not be able to be searched locally.

  -Q, --quiet-skip
    Do not warn when skipping download of already existing files.

  -O, --overwrite
    Force download and overwrite any files that already exist.


  --print-config-path
    Show the configuration file path.
    If the file doesn't exist, a default one is automatically copied.

  -C PATH, --config PATH
    Use `PATH` as configuration file instead of the default location.

  --help-order-values
    Show possible values for `-o`/`--order`.

  -h, --help
    Show this help.

  -V, --version
    Show the program version.

Notes:
  Pixiv ugoiras
    The converted webm seen in the browser will be downloaded,
    instead of the heavy and unplayable zip files normally provided by the
    Danbooru API.

  Info JSON keys
    The info returned for posts contains `fetched_at` (date) and
    `fetched_from` (booru name) keys not present in the standard API returns.

    Posts returned from local directories will not have the
    `keeper_data` and `pixiv_ugoira_frame_data` keys, they are not indexed
    for performance reasons.

  Decensooru
    Censored posts from Danbooru/Safebooru (banned artists and tags) are
    automatically fixed when possible, using Decensooru's database.

Examples:
  lunafind "blonde 2girls" --limit 200 --pages all --download .
  ⋅                        -l          -p          -D
    Download all pages of posts containing tags `blonde` and `2girls` to the
    current directory. Don't forget the quotes for tags.
    See the `-l`/`--limit` option description to know why `200` is used here.

  lunafind https://safebooru.donmai.us/posts/1667182 --query-location
  ⋅                                                  -q
    Print info for the post at this URL.
    If it was a downloaded post, its directory should be used instead of URL.

  lunafind id:1667182 --query-location
  ⋅                   -q
    Print info for post with ID 2092692 (same result as URL example above).

  lunafind "blonde 2girls" --source . --show-location media
  ⋅                        -s         -S
    Print image/webm path of all posts with the tags `blonde` and `2girls`
    downloaded in the current directory.

  lunafind --show-location post
  ⋅        -S
    Print URL of the latest posts on the home page.

  lunafind translated --resource notes
  ⋅                   -R
    Print notes JSON for latest posts with the tag `translated`.

  lunafind --random --limit 100 | jq .id
  ⋅        -r       -l
    Print the ID of 100 random posts,
    using `jq` to display the `id` key of each info JSON.

  lunafind wallpaper --pages all --filter "%-no_human ratio:16:9 width:>=1920"
  ⋅                  -p          -f
    Retrieve all posts with the `wallpaper` tags,
    filter them to only leave those without the `no_human` tag, with a ratio
    of 16:9 and a width equal or superior to 1920, print info.

    Since the filter value starts with a `-`, it is escaped with a `%` to not
    be mistaken for an option. `\` can also be used, but will most likely
    always require quoting due to the shell.

  lunafind "~scenery ~landscape" "~outdoor ~nature" --pages 1-10 --download .
  ⋅                                                 -p           -D
    Do two separate searches (because of Danbooru two tag limit) for
    "scenery OR landscape" and "outdoor OR nature", pages 1 to 10,
    combine the results and download everything."""

import re
import sys
from pathlib import Path
from types import GeneratorType
from typing import List, Optional

import docopt
from colorama import Fore

from . import LOG, Album, Stream, __about__, config, order, utils

OPTIONS = [string for match in re.findall(r"(-.)(?:\s|,)|(--.+?)\s", __doc__)
           for string in match if string]


def print_order_values() -> None:
    dicts     = {**order.ORDER_NUM, **order.ORDER_DATE, **order.ORDER_FUNCS}
    by_maxlen = len(max(dicts.keys(), key=len))

    for di in (order.ORDER_NUM, order.ORDER_DATE):
        print(f"{Fore.MAGENTA}{'Value':{by_maxlen}}  Default sort{Fore.RESET}")

        for by, (asc_or_desc, _) in di.items():
            print(f"{Fore.BLUE}{by:{by_maxlen}}{Fore.RESET}  "
                  f"{asc_or_desc}ending")

        print()

    print("For values above, the sort order can be manually specified by "
          "prefixing the value with 'asc_' or 'desc_', e.g. 'asc_score'.\n")

    for by in order.ORDER_FUNCS:
        print(f"{Fore.BLUE}{by}{Fore.RESET}")

    sys.exit(0)


def main(argv: Optional[List[str]] = None) -> None:
    argv = argv if argv is not None else sys.argv[1:]

    try:
        args = docopt.docopt(
            __doc__, help=False, argv=argv, version=__about__.__version__
        )
    except docopt.DocoptExit:
        LOG.error(f"Invalid command syntax or bad option, check --help.")
        sys.exit(10)

    if args["--config"]:
        config.FILE = Path(args["--config"])
        config.reload()

    if args["--help-order-values"]:
        print_order_values()

    if args["--help"]:
        utils.print_colored_help(__doc__)

    if args["--print-config-path"]:
        print(config.FILE)
        sys.exit()


    params = {
        "pages":  args["--pages"],
        "limit":  int(args["--limit"]) if args["--limit"] else None,
        "random": args["--random"],
        "raw":    args["--raw"],
        "client": args["--source"],
    }

    params = {k: v for k, v in params.items() if v is not None}

    unesc = lambda s: s[1:] if s.startswith(r"\-") or s.startswith("%-") else s

    stores = [
        Stream(unesc(q),
               **params,
               location     = args["--query-location"],
               partial_tags = args["--partial-match"])
        .filter(unesc(args["--filter"] or ""),
                partial_tags = args["--partial-match"])
        for q in args["QUERY"] or ("",)
    ]

    if args["--order"]:
        stores = [
            sum([Album(s) for s in stores], Album()).order(args["--order"])
        ]


    if not (args["--resource"] or args["--show-location"] or
            args["--download"]):
        args["--resource"] = "info"

    for obj in stores:
        posts = obj.list if isinstance(obj, Album) else obj

        if args["--download"]:
            posts.download(base_dir  = args["--download"],
                           overwrite = args["--overwrite"],
                           warn      = not args["--quiet-skip"])
            continue

        try:
            for post in posts:
                if args["--show-location"]:
                    path = post.get_location(
                        args["--show-location"],
                        absolute = args["--absolute-path"]
                    )
                    if path:
                        print(path, flush=True)
                    continue

                res_name = args["--resource"]
                res      = getattr(post, res_name) if res_name else post.info

                if res_name == "media" and isinstance(res, GeneratorType):
                    res = b"".join(res)

                if res and isinstance(res, bytes):
                    sys.stdout.buffer.write(res)
                    sys.stdout.flush()
                elif res:
                    print(utils.jsonify(res, indent=4), flush=True)
                else:
                    LOG.warning("Post %d has no %s resource.",
                                post.id, res_name)
        except (KeyboardInterrupt, BrokenPipeError):
            sys.exit(130)
